/*
 * submit_command_sync.c
 *
 *  Created on: 2 juil. 2015
 *      Author: cresson
 */

#include "drmaa.h"
#include <stdlib.h>     /* exit, EXIT_FAILURE */
#include <stdlib.h>
#include <string.h>

void concat(char *s1, char *s2)
{
    char *result = malloc(strlen(s1)+strlen(s2)+1);//+1 for the zero-terminator
    //in real code you would check for errors in malloc here
    strcpy(result, s1);
    strcat(result, s2);
free(s1);
    s1 = result;
}

int main (int argc, char **argv) {
if (argc<2)
{
printf ("The given command is empty!\n");
return EXIT_FAILURE;
}
   int rc = EXIT_SUCCESS;
   char error[DRMAA_ERROR_STRING_BUFFER];
   int errnum = 0;
   drmaa_job_template_t *jt = NULL;

   // Init. drmaa session
   errnum = drmaa_init (NULL, error, DRMAA_ERROR_STRING_BUFFER);

   if (errnum != DRMAA_ERRNO_SUCCESS) {
      fprintf (stderr, "Could not initialize the DRMAA library: %s\n", error);
      return EXIT_FAILURE;
   }

   // Create a job template
   errnum = drmaa_allocate_job_template (&jt, error, DRMAA_ERROR_STRING_BUFFER);

   if (errnum != DRMAA_ERRNO_SUCCESS) {
      fprintf (stderr, "Could not create job template: %s\n", error);
      rc = EXIT_FAILURE;
   }
   else {

      // Setting job attributes
      int i;
      int count = 0;
      for (i=1; i<argc; i++)
      {
        count += strlen(argv[i]); // argument lenght
        count += 1; // space between arguments
      }
      char command[count];
      strcpy(command, argv[1]);
      for (i=2; i<argc; i++)
      {
        strcat(command," ");
        strcat(command,argv[i]);
      }

      errnum = drmaa_set_attribute (jt, DRMAA_REMOTE_COMMAND, command /*"/gpfs-dell/data/work/GEOSUD/drmaa/test/drmaa_operator/bin/./sleeper.sh"*/,
                                    error, DRMAA_ERROR_STRING_BUFFER);

      if (errnum != DRMAA_ERRNO_SUCCESS) {
         fprintf (stderr, "Could not set attribute \"%s\": %s\n",
                  DRMAA_REMOTE_COMMAND, error);
         rc = EXIT_FAILURE;
      }
      else {
         // Submit job
         char jobid[DRMAA_JOBNAME_BUFFER];
         errnum = drmaa_run_job (jobid, DRMAA_JOBNAME_BUFFER, jt, error,
                                 DRMAA_ERROR_STRING_BUFFER);

         if (errnum != DRMAA_ERRNO_SUCCESS) {
            fprintf (stderr, "Could not submit job: %s\n", error);
            rc = EXIT_FAILURE;
         }
         else {
            printf ("Your job has been submitted with id %s\n", jobid);

            // wait job
            int status = 0;
            char jobid_out[DRMAA_JOBNAME_BUFFER];
            drmaa_attr_values_t *rusage = NULL;
            errnum = drmaa_wait (jobid, jobid_out, DRMAA_JOBNAME_BUFFER, &status,
                                 DRMAA_TIMEOUT_WAIT_FOREVER, &rusage, error,
                                 DRMAA_ERROR_STRING_BUFFER);

            if (errnum != DRMAA_ERRNO_SUCCESS) {
               fprintf (stderr, "Could not wait for job: %s\n", error);
            }
            else {
               char usage[DRMAA_ERROR_STRING_BUFFER];
               int aborted = 0;

               drmaa_wifaborted(&aborted, status, NULL, 0);

               if (aborted == 1) {
                  printf("Job %s never ran\n", jobid);
               }
               else {
                  int exited = 0;

                  drmaa_wifexited(&exited, status, NULL, 0);

                  if (exited == 1) {
                     int exit_status = 0;

                     drmaa_wexitstatus(&exit_status, status, NULL, 0);
                     printf("Job %s finished regularly with exit status %d\n", jobid, exit_status);
                  }
                  else {
                     int signaled = 0;

                     drmaa_wifsignaled(&signaled, status, NULL, 0);

                     if (signaled == 1) {
                        char termsig[DRMAA_SIGNAL_BUFFER+1];

                        drmaa_wtermsig(termsig, DRMAA_SIGNAL_BUFFER, status, NULL, 0);
                        printf("Job %s finished due to signal %s\n", jobid, termsig);
                     }
                     else {
                        printf("Job %s finished with unclear conditions\n", jobid);
                     }
                  } /* else */
               } /* else */

               printf ("Job Usage:\n");

               while (drmaa_get_next_attr_value (rusage, usage, DRMAA_ERROR_STRING_BUFFER) == DRMAA_ERRNO_SUCCESS) {
                  printf ("  %s\n", usage);
               }

               drmaa_release_attr_values (rusage);
            }
         }
      } /* else */

      // Delete job template
      errnum = drmaa_delete_job_template (jt, error, DRMAA_ERROR_STRING_BUFFER);

      if (errnum != DRMAA_ERRNO_SUCCESS) {
         fprintf (stderr, "Could not delete job template: %s\n", error);
         return EXIT_FAILURE;
      }
   } /* else */

   // Close drmaa session
   errnum = drmaa_exit (error, DRMAA_ERROR_STRING_BUFFER);

   if (errnum != DRMAA_ERRNO_SUCCESS) {
      fprintf (stderr, "Could not shut down the DRMAA library: %s\n", error);
      rc = EXIT_FAILURE;
   }

   return rc;
}






